import datetime
import json
from typing import Any, List

from hyfi.composer import BaseModel

from hyabsa import HyFI
from hyabsa.llms import ChatCompletionResponse

logger = HyFI.getLogger(__name__)


class AgentResult(BaseModel):
    timestamp: str = f"{datetime.datetime.now():%Y-%m-%d %H:%M:%S}"
    parsed: str
    usage: dict
    response: List[Any]

    @classmethod
    def from_chat_reponse(
        cls,
        response: ChatCompletionResponse,
    ) -> "AgentResult":
        parsed = "success"
        try:
            content = json.loads(response.content)
            if isinstance(content, dict):
                parsed = "failed"
                content = [content]
        except json.decoder.JSONDecodeError:
            content = [response.content]
            parsed = "failed"

        return cls(
            parsed=parsed,
            usage=response.usage,
            response=content,
        )



import datetime
import json
from typing import Optional

from hyfi.composer import BaseModel

from hyabsa import HyFI
from hyabsa.contexts import ChatMessage
from hyabsa.llms import OpenAIChatCompletion
from hyabsa.prompts import Prompts

from .results import AgentResult

logger = HyFI.getLogger(__name__)


class BaseAgent(BaseModel):
    _config_group_: str = "/agent"
    _config_name_: str = "__init__"

    llm: OpenAIChatCompletion = OpenAIChatCompletion()
    prompts: Prompts = Prompts()
    task: str = "AE"
    prompt: str = "base"
    output_dir: str = "outputs/preds"
    output_filename: Optional[str] = None
    verbose: bool = False

    @property
    def model(self) -> str:
        return self.llm.model.replace(".", "")

    @property
    def output_filepath(self) -> str:
        ouput_filename = (
            self.output_filename
            or f"{self.task}_{self.prompt}_{self.model}_{datetime.datetime.now().strftime('%Y%m%d')}.jsonl"
        )
        HyFI.mkdir(self.output_dir)
        return f"{self.output_dir}/{ouput_filename}"

    def build_message(self, text: str) -> ChatMessage:
        return ChatMessage(content=self.prompts.build(text, self.task, self.prompt))

    def execute(self, text: str) -> str:
        result = AgentResult.from_chat_reponse(
            self.llm.request(self.build_message(text)),
        )
        if self.output_filepath:
            HyFI.append_to_jsonl(result.model_dump(), self.output_filepath)
        return json.dumps(result.model_dump(), ensure_ascii=False)



import json

from hyabsa import HyFI
from hyabsa.llms import OpenAIChatCompletion

from .base import BaseAgent
from .results import AgentResult

logger = HyFI.getLogger(__name__)


class AbsaAgent(BaseAgent):
    _config_name_: str = "absa"
    llm: OpenAIChatCompletion = OpenAIChatCompletion()

    def execute(self, text: str) -> str:
        msg = self.build_message(text)
        rst = self.llm.request(msg)
        result = AgentResult.from_chat_reponse(rst)
        if self.output_filepath:
            HyFI.append_to_jsonl(result.model_dump(), self.output_filepath)
        return json.dumps(result.model_dump(), ensure_ascii=False)



import time
from typing import Any, Dict, List, Optional

import backoff
import openai
from hyfi.composer import BaseModel
from openai.error import (
    APIConnectionError,
    APIError,
    InvalidRequestError,
    RateLimitError,
    ServiceUnavailableError,
)

from hyabsa import HyFI
from hyabsa.contexts import ChatMessage

logger = HyFI.getLogger(__name__)


class ChatCompletionConfig(BaseModel):
    model: str
    temperature: float
    messages: List[ChatMessage]


class ChatCompletionResponse(BaseModel):
    content: str
    usage: Dict[str, Any] = {
        "prompt_tokens": 0,
        "completion_tokens": 0,
        "total_tokens": 0,
    }


class OpenAIChatCompletion(BaseModel):
    _config_group_: str = "/llm"
    _config_name_: str = "openai"

    api_key: Optional[str] = None
    model: str = "gpt-3.5-turbo-0301"
    rate_limit_per_minute: int = 3500
    temperature: float = 0.0

    _engine_: Optional[openai.ChatCompletion] = None

    @property
    def engine(self):
        if self._engine_ is None:
            self.initialize()
        return self._engine_

    def initialize(self, api_key: Optional[str] = None):
        api_key = api_key or self.api_key
        denv = HyFI.DotEnvConfig()
        if not api_key and denv.OPENAI_API_KEY:
            api_key = denv.OPENAI_API_KEY.get_secret_value()
        if not api_key:
            raise ValueError("OpenAI API Key is required.")
        openai.api_key = api_key
        logger.info("OpenAI API Key is set successfully.")
        self._engine_ = openai.ChatCompletion()
        logger.info("OpenAI ChatCompletion API is initialized.")

    def build_config(self, message: ChatMessage) -> Dict[str, Any]:
        return ChatCompletionConfig(
            model=self.model,
            temperature=self.temperature,
            messages=[message],
        ).model_dump()

    def request(self, message: ChatMessage) -> ChatCompletionResponse:
        delay = 60.0 / self.rate_limit_per_minute
        return call_api(
            self.engine,
            self.build_config(message),
            delay_in_seconds=delay,
        )


@backoff.on_exception(
    backoff.expo,
    (
        RateLimitError,
        APIConnectionError,
        APIError,
        ServiceUnavailableError,
    ),
)
def request_api(engine, args, delay_in_seconds: float = 1):
    time.sleep(delay_in_seconds)
    return engine.create(**args)


def call_api(engine, args, delay_in_seconds: float = 1) -> ChatCompletionResponse:
    time.sleep(delay_in_seconds)
    try:
        response = request_api(engine, args, delay_in_seconds=delay_in_seconds)
        message = response["choices"][0]["message"]
        content = message["content"].strip().strip("\n")
        usage = response["usage"]
        return ChatCompletionResponse(content=content, usage=usage)
    except InvalidRequestError as e:
        logger.error(e)
        return ChatCompletionResponse(content=e.user_message)
    except Exception as e:
        logger.error(e)
        return ChatCompletionResponse(content=str(e))



from typing import List, Optional, Union

from datasets import Dataset
from hyfi.run import RunConfig
from hyfi.runner import BaseRunner

from hyabsa import HyFI
from hyabsa.agents import AbsaAgent

logger = HyFI.getLogger(__name__)


class AbsaRunner(BaseRunner):
    _config_name_: str = "absa"
    _auto_populate_: bool = True

    task_name: str = "absa"
    agent: AbsaAgent = AbsaAgent()
    tasks: Optional[List[str]] = []
    data_load: RunConfig = RunConfig(_config_name_="load_dataset")
    data_save: RunConfig = RunConfig(_config_name_="save_dataset_to_disk")

    text_col: str = "bodyText"
    batch_size: int = 1000
    num_workers: int = 1
    remove_columns: Optional[Union[List[str], str]] = None
    load_from_cache_file: bool = True
    top_n: Optional[int] = None

    _dataset: Optional[Dataset] = None

    @property
    def dataset(self) -> Dataset:
        if self._dataset is None:
            self._dataset = self.load_dataset()
        return self._dataset

    def load_dataset(self) -> Dataset:
        return HyFI.partial(self.data_load.config)()

    def save_dataset(self, dataset: Dataset) -> None:
        HyFI.partial(self.data_save.config)(dataset)

    def run(self):
        dataset = self.load_dataset()
        if self.top_n is not None:
            logger.info("Selecting top %s rows...", self.top_n)
            dataset = dataset.select(range(self.top_n))

        tasks = self.tasks or ["QUAD"]
        for task in tasks:
            logger.info(
                "Predicting %s with a batch size of %s and %s workers...",
                task,
                self.batch_size,
                self.num_workers,
            )
            dataset = dataset.map(
                batch_run,
                batched=True,
                batch_size=self.batch_size,
                num_proc=self.num_workers,
                fn_kwargs={
                    "task": task,
                    "agent": self.agent.model_copy(),
                    "text_col": self.text_col,
                },
                remove_columns=self.remove_columns,
                load_from_cache_file=self.load_from_cache_file,
            )
        if self.verbose:
            logger.info("Dataset shape: %s", dataset.shape)
            logger.info("Dataset columns: %s", dataset.column_names)
            print(dataset[0])
        self.save_dataset(dataset)


def batch_run(
    batch,
    task: str = "QUAD",
    agent: Optional[AbsaAgent] = None,
    text_col: str = "text",
) -> dict:
    if agent is None:
        raise ValueError("Agent must be provided")
    agent.task = task
    if agent.verbose:
        HyFI.print(agent.model_dump())

    res = [execute_each(text, agent) for text in batch[text_col]]
    return {f"{task}_pred": res}


def execute_each(
    text: str,
    agent: AbsaAgent,
) -> str:
    response = agent.execute(text)
    if agent.verbose:
        # logger.info("Text: %s", text)
        logger.info("Response: %s", response)

    return response



from typing import Dict

from hyfi.composer import BaseModel

from hyabsa import HyFI

logger = HyFI.getLogger(__name__)


class Prompts(BaseModel):
    _config_group_: str = "/prompts"
    _config_name_: str = "__init__"

    tasks: Dict[str, str] = {}
    prompts: Dict[str, Dict[str, str]] = {}

    def _get_prompt(self, task: str, prompt_name: str) -> str:
        if prompt := self.prompts.get(task, {}).get(prompt_name, ""):
            return prompt
        else:
            raise ValueError(f"Prompt for task {task} is not defined.")

    def build(self, text: str, task: str, prompt_name: str) -> str:
        prompt = self._get_prompt(task, prompt_name)
        prompt += f'\nInput text:\n"{text}"\nAnswer:\n'
        return prompt
